# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# source: https://github.com/yhhhli/BRECQ/tree/main/quant

import torch
import torch.distributed as dist
import torch.nn.functional as F
from typing import Union
from tqdm import tqdm
from trailmet.algorithms.quantize.quantize import (
    StraightThrough,
    BaseQuantization as BQ,
)
from trailmet.algorithms.quantize.qmodel import QuantModule, BaseQuantBlock
from trailmet.algorithms.quantize.methods import AdaRoundQuantizer
from trailmet.utils import lp_loss

__all__ = [
    'StopForwardException',
    'DataSaverHook',
    'GetLayerInpOut',
    'save_inp_oup_data',
    'GradSaverHook',
    'GetLayerGrad',
    'save_grad_data',
    'LinearTempDecay',
    'LayerLossFunction',
    'layer_reconstruction',
    'BlockLossFunction',
    'block_reconstruction',
]

optimizer_map = {
    'sgd': torch.optim.SGD,
    'adam': torch.optim.Adam,
    'adagrad': torch.optim.Adagrad,
    'adadelta': torch.optim.Adadelta,
}


class StopForwardException(Exception):
    """Used to throw and catch an exception to stop traversing the graph."""

    pass


class DataSaverHook:
    """Forward hook that stores the input and output of a layer."""

    def __init__(self,
                 store_input=False,
                 store_output=False,
                 stop_forward=False):
        self.store_input = store_input
        self.store_output = store_output
        self.stop_forward = stop_forward

        self.input_store = None
        self.output_store = None

    def __call__(self, module, input_batch, output_batch):
        if self.store_input:
            self.input_store = input_batch
        if self.store_output:
            self.output_store = output_batch
        if self.stop_forward:
            raise StopForwardException


class GetLayerInpOut:
    """Get the input and output of a specified layer in a quantized model.

    :param model: quantized model for which the input and output needs to be
        extracted.
    :param layer: the layer for which input and output needs to be extracted.
    :param device: the device on which the computation needs to be performed.
    :param asym: save quantized input and full precision output.
        [default=False]
    :param act_quant: use activation quantization. [default=False]
    """

    def __init__(
        self,
        model,
        layer: Union[QuantModule, BaseQuantBlock],
        device: torch.device,
        asym: bool = False,
        act_quant: bool = False,
    ):
        self.model = model
        self.layer = layer
        self.asym = asym
        self.device = device
        self.act_quant = act_quant
        self.data_saver = DataSaverHook(store_input=True,
                                        store_output=True,
                                        stop_forward=True)

    def __call__(self, model_input):
        """
        :param model_input: calibration data samples
        :return: tuple of layer input and output
        """
        self.model.eval()
        self.model.set_quant_state(False, False)

        handle = self.layer.register_forward_hook(self.data_saver)
        with torch.no_grad():
            try:
                _ = self.model(model_input.to(self.device))
            except StopForwardException:
                pass

            if self.asym:
                self.data_saver.store_output = False
                self.model.set_quant_state(weight_quant=True,
                                           act_quant=self.act_quant)
                try:
                    _ = self.model(model_input.to(self.device))
                except StopForwardException:
                    pass
                self.data_saver.store_output = True

        handle.remove()

        self.model.set_quant_state(False, False)
        self.layer.set_quant_state(True, self.act_quant)
        self.model.train()

        return (
            self.data_saver.input_store[0].detach(),
            self.data_saver.output_store.detach(),
        )


def save_inp_oup_data(
    model,
    layer: Union[QuantModule, BaseQuantBlock],
    cali_data: torch.Tensor,
    asym: bool = False,
    act_quant: bool = False,
    batch_size: int = 32,
    keep_gpu: bool = True,
):
    """Function to save input data and output data of a particular layer/block
    over calibration dataset.

    :param model: quantized model for which the input and output needs to be
        extracted.
    :param layer: the layer for which input and output needs to be extracted.
    :param cali_data: calibration dataset
    :param asym: save quantized input and full precision output.
        [default=False]
    :param act_quant: use activation quantization. [default=False]
    :param batch_size: mini-batch size for calibration. [default=32]
    :param keep_gpu: put saved data on GPU for faster optimization.
        [default=True]
    :return: input and output data
    """
    device = next(model.parameters()).device
    get_inp_out = GetLayerInpOut(model,
                                 layer,
                                 device=device,
                                 asym=asym,
                                 act_quant=act_quant)
    cached_batches = []
    torch.cuda.empty_cache()

    for i in range(int(cali_data.size(0) / batch_size)):
        cur_inp, cur_out = get_inp_out(cali_data[i * batch_size:(i + 1) *
                                                 batch_size])
        cached_batches.append((cur_inp.cpu(), cur_out.cpu()))

    cached_inps = torch.cat([x[0] for x in cached_batches])
    cached_outs = torch.cat([x[1] for x in cached_batches])
    torch.cuda.empty_cache()
    if keep_gpu:
        cached_inps = cached_inps.to(device)
        cached_outs = cached_outs.to(device)
    return cached_inps, cached_outs


class GradSaverHook:
    """Backward hook that stores the gradients of a layer."""

    def __init__(self, store_grad=True):
        self.store_grad = store_grad
        self.stop_backward = False
        self.grad_out = None

    def __call__(self, module, grad_input, grad_output):
        if self.store_grad:
            self.grad_out = grad_output[0]
        if self.stop_backward:
            raise StopForwardException


class GetLayerGrad:
    """Get the gradient a specified layer in a quantized model.

    :param model: quantized model for which the input and output needs to be
        extracted.
    :param layer: the layer for which input and output needs to be extracted.
    :param device: the device on which the computation needs to be performed.
    :param asym: if True, save quantized input and full precision output.
        [default=False]
    :param act_quant: use activation quantization. [default=False]
    """

    def __init__(
        self,
        model,
        layer: Union[QuantModule, BaseQuantBlock],
        device: torch.device,
        act_quant: bool = False,
    ):
        self.model = model
        self.layer = layer
        self.device = device
        self.act_quant = act_quant
        self.data_saver = GradSaverHook(True)

    def __call__(self, model_input):
        """Compute the gradients of layer output, note that we compute the
        gradient by calculating the KL loss between fp model and quant model.

        :param model_input: calibration data samples
        :return: gradients for the layer
        """
        self.model.eval()

        handle = self.layer.register_backward_hook(self.data_saver)
        with torch.enable_grad():
            try:
                self.model.zero_grad()
                inputs = model_input.to(self.device)
                self.model.set_quant_state(False, False)
                out_fp = self.model(inputs)
                self.model.quantize_model_till(self.layer, self.act_quant)
                out_q = self.model(inputs)
                loss = F.kl_div(
                    F.log_softmax(out_q, dim=1),
                    F.softmax(out_fp, dim=1),
                    reduction='batchmean',
                )
                loss.backward()
            except StopForwardException:
                pass

        handle.remove()
        self.model.set_quant_state(False, False)
        self.layer.set_quant_state(True, self.act_quant)
        self.model.train()
        return self.data_saver.grad_out.data


def save_grad_data(
    model,
    layer: Union[QuantModule, BaseQuantBlock],
    cali_data: torch.Tensor,
    damping: float = 1.0,
    act_quant: bool = False,
    batch_size: int = 32,
    keep_gpu: bool = True,
):
    """Function to save gradient data of a particular layer/block over
    calibration dataset.

    :param model: quantized model for which the input and output needs to be
        extracted.
    :param layer: the layer for which input and output needs to be extracted.
    :param cali_data: calibration dataset
    :param damping: damping the second-order gradient by adding some constant
        in the FIM diagonal
    :param act_quant: use activation quantization. [default=False]
    :param batch_size: mini-batch size for calibration. [default=32]
    :param keep_gpu: put saved data on GPU for faster optimization.
        [default=True]
    :return: gradient data
    """
    device = next(model.parameters()).device
    get_grad = GetLayerGrad(model, layer, device, act_quant=act_quant)
    cached_batches = []
    torch.cuda.empty_cache()

    for i in range(int(cali_data.size(0) / batch_size)):
        cur_grad = get_grad(cali_data[i * batch_size:(i + 1) * batch_size])
        cached_batches.append(cur_grad.cpu())

    cached_grads = torch.cat([x for x in cached_batches])
    cached_grads = cached_grads.abs() + 1.0
    # scaling to make sure its mean is 1
    # cached_grads = cached_grads * torch.sqrt(cached_grads.numel() / cached_grads.pow(2).sum())
    torch.cuda.empty_cache()
    if keep_gpu:
        cached_grads = cached_grads.to(device)
    return cached_grads


# ================================
# ****** Reconstruct Layer *******
# ================================


class LinearTempDecay:
    """Class to implement a linear temperature decay scheduler for a given
    maximum time step.

    :param t_max: maximum number of time steps to decay temperature over.
    :param rel_start_decay: relative point in time to start the decay from the
        maximum time step. [default=.2]
    :param start_b: initial temperature value. [default=10]
    :param end_b: final temperature value. [default=2]
    """

    def __init__(
        self,
        t_max: int,
        rel_start_decay: float = 0.2,
        start_b: int = 10,
        end_b: int = 2,
    ):
        self.t_max = t_max
        self.start_decay = rel_start_decay * t_max
        self.start_b = start_b
        self.end_b = end_b

    def __call__(self, t):
        """Cosine annealing scheduler for temperature b.

        :param t: the current time step
        :return: scheduled temperature
        """
        if t < self.start_decay:
            return self.start_b
        else:
            rel_t = (t - self.start_decay) / (self.t_max - self.start_decay)
            return self.end_b + (self.start_b - self.end_b) * max(
                0.0, (1 - rel_t))


class LayerLossFunction:

    def __init__(
        self,
        layer: QuantModule,
        round_loss: str = 'relaxation',
        weight: float = 1.0,
        rec_loss: str = 'mse',
        max_count: int = 2000,
        b_range: tuple = (10, 2),
        decay_start: float = 0.0,
        warmup: float = 0.0,
        p: float = 2.0,
    ):
        self.layer = layer
        self.round_loss = round_loss
        self.weight = weight
        self.rec_loss = rec_loss
        self.loss_start = max_count * warmup
        self.p = p
        self.count = 0
        # self.pbar = tqdm(total=max_count)
        self.pbar = tqdm(
            total=max_count,
            desc='Reconstructing Layer: Loss (X.X) b (X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
        )
        self.temp_decay = LinearTempDecay(
            max_count,
            rel_start_decay=warmup + (1 - warmup) * decay_start,
            start_b=b_range[0],
            end_b=b_range[1],
        )

    def __call__(self, pred, tgt, grad=None):
        """
        Compute the total loss for adaptive rounding:
        rec_loss is the quadratic output reconstruction loss, round_loss is
        a regularization term to optimize the rounding policy
        :param pred: output from quantized model
        :param tgt: output from FP model
        :param grad: gradients to compute fisher information
        :return: total loss function
        """
        self.count += 1
        if self.rec_loss == 'mse':
            rec_loss = lp_loss(pred, tgt, p=self.p)
        elif self.rec_loss == 'fisher_diag':
            rec_loss = ((pred - tgt).pow(2) * grad.pow(2)).sum(1).mean()
        elif self.rec_loss == 'fisher_full':
            a = (pred - tgt).abs()
            grad = grad.abs()
            batch_dotprod = torch.sum(a * grad, (1, 2, 3)).view(-1, 1, 1, 1)
            rec_loss = (batch_dotprod * a * grad).mean() / 100
        else:
            raise ValueError(
                'Not supported reconstruction loss function: {}'.format(
                    self.rec_loss))

        b = self.temp_decay(self.count)
        if self.count < self.loss_start or self.round_loss == 'none':
            b = round_loss = 0
        elif self.round_loss == 'relaxation':
            round_loss = 0
            round_vals = self.layer.weight_quantizer.get_soft_targets()
            round_loss += (self.weight *
                           (1 - ((round_vals - 0.5).abs() * 2).pow(b)).sum())
        else:
            raise NotImplementedError

        total_loss = rec_loss + round_loss

        if self.count % 100 == 0:
            self.pbar.set_description(
                'Reconstructing Layer: Loss ({:.3f}) b ({:.1f})'.format(
                    float(total_loss), b))
        #     self.pbar.set_postfix(loss=float(total_loss), b=b)
        self.pbar.update(1)
        return total_loss


def layer_reconstruction(
    model,
    layer: QuantModule,
    cali_data: torch.Tensor,
    batch_size: int = 32,
    iters: int = 20000,
    weight: float = 0.001,
    opt_mode: str = 'mse',
    asym: bool = False,
    include_act_func: bool = True,
    b_range: tuple = (20, 2),
    warmup: float = 0.0,
    act_quant: bool = False,
    lr: float = 4e-5,
    p: float = 2.0,
    multi_gpu: bool = False,
    optim='adam',
):
    """Block reconstruction to optimize the output from each layer.

    :param model: QuantModel
    :param layer: QuantModule that needs to be optimized
    :param cali_data: data for calibration, typically 1024 training images, as
        described in AdaRound
    :param batch_size: mini-batch size for reconstruction
    :param iters: optimization iterations for reconstruction,
    :param weight: the weight of rounding regularization term
    :param opt_mode: optimization mode
    :param asym: asymmetric optimization designed in AdaRound, use quant input
        to reconstruct fp output
    :param include_act_func: optimize the output after activation function
    :param b_range: temperature range
    :param warmup: proportion of iterations that no scheduling for temperature
    :param act_quant: use activation quantization or not.
    :param lr: learning rate for act delta learning
    :param p: L_p norm minimization
    :param multi_gpu: use multi-GPU or not, if enabled, we should sync the
        gradients
    """

    model.set_quant_state(False, False)
    layer.set_quant_state(True, act_quant)
    round_mode = 'learned_hard_sigmoid'

    if not include_act_func:
        org_act_func = layer.activation_function
        layer.activation_function = StraightThrough()

    if not act_quant:
        # Replace weight quantizer to AdaRoundQuantizer
        layer.weight_quantizer = AdaRoundQuantizer(
            uaq=layer.weight_quantizer,
            round_mode=round_mode,
            weight_tensor=layer.org_weight.data,
        )
        layer.weight_quantizer.soft_targets = True

        # Set up optimizer
        opt_params = [layer.weight_quantizer.alpha]
        optimizer = optimizer_map[optim](opt_params)
        scheduler = None
    else:
        # Use UniformAffineQuantizer to learn delta
        opt_params = [layer.act_quantizer.delta]
        optimizer = optimizer_map[optim](opt_params, lr=lr)
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer,
                                                               T_max=iters,
                                                               eta_min=0.0)

    loss_mode = 'none' if act_quant else 'relaxation'
    rec_loss = opt_mode

    loss_func = LayerLossFunction(
        layer,
        round_loss=loss_mode,
        weight=weight,
        max_count=iters,
        rec_loss=rec_loss,
        b_range=b_range,
        decay_start=0,
        warmup=warmup,
        p=p,
    )

    # Save data before optimizing the rounding
    cached_inps, cached_outs = save_inp_oup_data(model, layer, cali_data, asym,
                                                 act_quant, batch_size)
    if opt_mode != 'mse':
        cached_grads = save_grad_data(model,
                                      layer,
                                      cali_data,
                                      act_quant,
                                      batch_size=batch_size)
    else:
        cached_grads = None
    device = 'cuda'
    for i in range(iters):
        idx = torch.randperm(cached_inps.size(0))[:batch_size]
        cur_inp = cached_inps[idx]
        cur_out = cached_outs[idx]
        cur_grad = cached_grads[idx] if opt_mode != 'mse' else None

        optimizer.zero_grad()
        out_quant = layer(cur_inp)

        err = loss_func(out_quant, cur_out, cur_grad)
        err.backward(retain_graph=True)
        if multi_gpu:
            for p in opt_params:
                dist.all_reduce(p.grad)
        optimizer.step()
        if scheduler:
            scheduler.step()

    torch.cuda.empty_cache()

    # Finish optimization, use hard rounding.
    layer.weight_quantizer.soft_targets = False

    # Reset original activation function
    if not include_act_func:
        layer.activation_function = org_act_func


# =================================
# ******* Reconstruct Block *******
# =================================


class BlockLossFunction:

    def __init__(
        self,
        block: BaseQuantBlock,
        round_loss: str = 'relaxation',
        weight: float = 1.0,
        rec_loss: str = 'mse',
        max_count: int = 2000,
        b_range: tuple = (10, 2),
        decay_start: float = 0.0,
        warmup: float = 0.0,
        p: float = 2.0,
    ):
        self.block = block
        self.round_loss = round_loss
        self.weight = weight
        self.rec_loss = rec_loss
        self.loss_start = max_count * warmup
        self.p = p
        self.count = 0
        # self.pbar = tqdm(total=max_count)
        self.pbar = tqdm(
            total=max_count,
            desc='Reconstructing Block: Loss (X.X) b (X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
        )
        self.temp_decay = LinearTempDecay(
            max_count,
            rel_start_decay=warmup + (1 - warmup) * decay_start,
            start_b=b_range[0],
            end_b=b_range[1],
        )

    def __call__(self, pred, tgt, grad=None):
        """
        Compute the total loss for adaptive rounding:
        rec_loss is the quadratic output reconstruction loss, round_loss is
        a regularization term to optimize the rounding policy
        :param pred: output from quantized model
        :param tgt: output from FP model
        :param grad: gradients to compute fisher information
        :return: total loss function
        """
        self.count += 1
        if self.rec_loss == 'mse':
            rec_loss = lp_loss(pred, tgt, p=self.p)
        elif self.rec_loss == 'fisher_diag':
            rec_loss = ((pred - tgt).pow(2) * grad.pow(2)).sum(1).mean()
        elif self.rec_loss == 'fisher_full':
            a = (pred - tgt).abs()
            grad = grad.abs()
            batch_dotprod = torch.sum(a * grad, (1, 2, 3)).view(-1, 1, 1, 1)
            rec_loss = (batch_dotprod * a * grad).mean() / 100
        else:
            raise ValueError(
                'Not supported reconstruction loss function: {}'.format(
                    self.rec_loss))

        b = self.temp_decay(self.count)
        if self.count < self.loss_start or self.round_loss == 'none':
            b = round_loss = 0
        elif self.round_loss == 'relaxation':
            round_loss = 0
            for name, module in self.block.named_modules():
                if isinstance(module, QuantModule):
                    round_vals = module.weight_quantizer.get_soft_targets()
                    round_loss += (self.weight * (1 - (
                        (round_vals - 0.5).abs() * 2).pow(b)).sum())
        else:
            raise NotImplementedError

        total_loss = rec_loss + round_loss
        if self.count % 100 == 0:
            self.pbar.set_description(
                'Reconstructing Block: Loss ({:.3f}) b ({:.1f})'.format(
                    float(total_loss), b))
        #     self.pbar.set_postfix(loss=float(total_loss), b=b)
        self.pbar.update(1)
        return total_loss


def block_reconstruction(
    model,
    block: BaseQuantBlock,
    cali_data: torch.Tensor,
    batch_size: int = 32,
    iters: int = 20000,
    weight: float = 0.01,
    opt_mode: str = 'mse',
    asym: bool = False,
    include_act_func: bool = True,
    b_range: tuple = (20, 2),
    warmup: float = 0.0,
    act_quant: bool = False,
    lr: float = 4e-5,
    p: float = 2.0,
    multi_gpu: bool = False,
    optim='adam',
):
    """Block reconstruction to optimize the output from each block.

    :param model: QuantModel
    :param block: BaseQuantBlock that needs to be optimized
    :param cali_data: data for calibration, typically 1024 training images, as
        described in AdaRound
    :param batch_size: mini-batch size for reconstruction
    :param iters: optimization iterations for reconstruction,
    :param weight: the weight of rounding regularization term
    :param opt_mode: optimization mode
    :param asym: asymmetric optimization designed in AdaRound, use quant input
        to reconstruct fp output
    :param include_act_func: optimize the output after activation function
    :param b_range: temperature range
    :param warmup: proportion of iterations that no scheduling for temperature
    :param act_quant: use activation quantization or not.
    :param lr: learning rate for act delta learning
    :param p: L_p norm minimization
    :param multi_gpu: use multi-GPU or not, if enabled, we should sync the
        gradients
    """
    model.set_quant_state(False, False)
    block.set_quant_state(True, act_quant)
    round_mode = 'learned_hard_sigmoid'

    if not include_act_func:
        org_act_func = block.activation_function
        block.activation_function = StraightThrough()

    if not act_quant:
        # Replace weight quantizer to AdaRoundQuantizer
        for name, module in block.named_modules():
            if isinstance(module, QuantModule):
                module.weight_quantizer = AdaRoundQuantizer(
                    uaq=module.weight_quantizer,
                    round_mode=round_mode,
                    weight_tensor=module.org_weight.data,
                )
                module.weight_quantizer.soft_targets = True

        # Set up optimizer
        opt_params = []
        for name, module in block.named_modules():
            if isinstance(module, QuantModule):
                opt_params += [module.weight_quantizer.alpha]
        optimizer = optimizer_map[optim](opt_params)
        scheduler = None
    else:
        # Use UniformAffineQuantizer to learn delta
        if hasattr(block.act_quantizer, 'delta'):
            opt_params = [block.act_quantizer.delta]
        else:
            opt_params = []
        for name, module in block.named_modules():
            if isinstance(module, QuantModule):
                if module.act_quantizer.delta is not None:
                    opt_params += [module.act_quantizer.delta]
        optimizer = optimizer_map[optim](opt_params, lr=lr)
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer,
                                                               T_max=iters,
                                                               eta_min=0.0)

    loss_mode = 'none' if act_quant else 'relaxation'
    rec_loss = opt_mode

    loss_func = BlockLossFunction(
        block,
        round_loss=loss_mode,
        weight=weight,
        max_count=iters,
        rec_loss=rec_loss,
        b_range=b_range,
        decay_start=0,
        warmup=warmup,
        p=p,
    )

    # Save data before optimizing the rounding
    cached_inps, cached_outs = save_inp_oup_data(model, block, cali_data, asym,
                                                 act_quant, batch_size)
    if opt_mode != 'mse':
        cached_grads = save_grad_data(model,
                                      block,
                                      cali_data,
                                      act_quant,
                                      batch_size=batch_size)
    else:
        cached_grads = None
    device = 'cuda'
    for i in range(iters):
        idx = torch.randperm(cached_inps.size(0))[:batch_size]
        cur_inp = cached_inps[idx].to(device)
        cur_out = cached_outs[idx].to(device)
        cur_grad = cached_grads[idx].to(device) if opt_mode != 'mse' else None

        optimizer.zero_grad()
        out_quant = block(cur_inp)

        err = loss_func(out_quant, cur_out, cur_grad)
        err.backward(retain_graph=True)
        if multi_gpu:
            for p in opt_params:
                dist.all_reduce(p.grad)
        optimizer.step()
        if scheduler:
            scheduler.step()

    torch.cuda.empty_cache()

    # Finish optimization, use hard rounding.
    for name, module in block.named_modules():
        if isinstance(module, QuantModule):
            module.weight_quantizer.soft_targets = False

    # Reset original activation function
    if not include_act_func:
        block.activation_function = org_act_func



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import sys
import time
import torch
import torch.nn as nn
import torch.utils
import torch.backends.cudnn as cudnn
import torch.utils.data.distributed
from trailmet.algorithms.binarize.binarize import BaseBinarize
from trailmet.utils import (
    AverageMeter,
    save_checkpoint,
    accuracy,
    CrossEntropyLabelSmooth,
)

import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import numpy as np

logger = logging.getLogger(__name__)


class BirealNet(BaseBinarize):

    def __init__(self, model, dataloaders, **CFG):
        self.model = model
        self.dataloaders = dataloaders
        self.CFG = CFG
        self.batch_size = self.CFG['batch_size']
        self.optimizer = self.CFG['optimizer']
        self.epochs = self.CFG['epochs']
        self.lr = self.CFG['lr']
        self.momentum = self.CFG['momentum']
        self.save_path = self.CFG['save_path']
        self.data_path = self.CFG['data_path']
        self.weight_decay = self.CFG['weight_decay']
        self.label_smooth = self.CFG['label_smooth']
        self.num_workers = self.CFG['workers']
        self.dataset = self.CFG['dataset']
        self.num_class = self.CFG['num_classes']
        self.device = self.CFG['device']

        self.wandb_monitor = self.CFG.get('wandb', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__

        self.name = '_'.join([
            self.dataset_name,
            f'{self.epochs}',
            f'{self.lr}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/BiRealNet', exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/BiRealNet/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.CFG}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet BiRealNet', name=self.name)
            wandb.config.update(self.CFG)

    def train(self, epoch, train_loader, model, criterion, optimizer,
              scheduler):
        batch_time = AverageMeter('Time', ':6.3f')
        data_time = AverageMeter('Data', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            train_loader,
            desc=
            'Training BiRealNet Epoch [X] (X / X Steps) (batch time=X.Xs) (data time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        model.train()
        end = time.time()
        scheduler.step()

        for param_group in optimizer.param_groups:
            cur_lr = param_group['lr']
        print('learning_rate:', cur_lr)

        for i, (images, target) in enumerate(epoch_iterator):
            data_time.update(time.time() - end)
            images = images.to(device=self.device)
            target = target.to(device=self.device)

            # compute outputy
            logits = model(images)
            loss = criterion(logits, target)

            # measure accuracy and record loss
            prec1, prec5 = accuracy(logits, target, topk=(1, 5))
            n = images.size(0)
            losses.update(loss.item(), n)  # accumulated loss
            top1.update(prec1.item(), n)
            top5.update(prec5.item(), n)

            # compute gradient and do SGD step
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            epoch_iterator.set_description(
                'Training BiRealNet Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            logger.info(
                'Training BiRealNet Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(train_loader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                    top1.val,
                    top5.val,
                ))

            if self.wandb_monitor:
                wandb.log({
                    'train_loss': losses.val,
                    'train_top1_acc': top1.val,
                    'train_top5_acc': top5.val,
                })

        return losses.avg, top1.avg, top5.avg

    def validate(self, epoch, val_loader, model, criterion, CFG):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            val_loader,
            desc=
            'Validating BiRealNet Epoch [X] (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        # switch to evaluation mode
        model.eval()
        with torch.no_grad():
            end = time.time()
            for i, (images, target) in enumerate(epoch_iterator):
                images = images.to(device=self.device)
                target = target.to(device=self.device)

                # compute output
                logits = model(images)
                loss = criterion(logits, target)

                # measure accuracy and record loss
                pred1, pred5 = accuracy(logits, target, topk=(1, 5))
                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating BiRealNet Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(val_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating BiRealNet Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(val_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'val_loss': losses.val,
                        'val_top1_acc': top1.val,
                        'val_top5_acc': top5.val,
                    })

            print(
                ' * Val acc@1 {top1.avg:.3f} Val acc@5 {top5.avg:.3f}'.format(
                    top1=top1, top5=top5))

        return losses.avg, top1.avg, top5.avg

    def test(self, epoch, test_loader, model, criterion, CFG):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            test_loader,
            desc=
            'Testing BiRealNet Epoch [X] (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        # switch to evaluation mode
        model.eval()
        with torch.no_grad():
            end = time.time()
            for i, (images, target) in enumerate(epoch_iterator):
                images = images.to(device=self.device)
                target = target.to(device=self.device)

                # compute output
                logits = model(images)
                loss = criterion(logits, target)

                # measure accuracy and record loss
                pred1, pred5 = accuracy(logits, target, topk=(1, 5))
                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Testing BiRealNet Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(test_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Testing BiRealNet Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(test_loader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'test_loss': losses.val,
                        'test_top1_acc': top1.val,
                        'test_top5_acc': top5.val,
                    })

            print(' * Test acc@1 {top1.avg:.3f} Test acc@5 {top5.avg:.3f}'.
                  format(top1=top1, top5=top5))

        return losses.avg, top1.avg, top5.avg

    def binarize(self):
        if not torch.cuda.is_available():
            sys.exit(1)
        start_t = time.time()

        cudnn.benchmark = True
        cudnn.enabled = True

        # load model
        model = self.model
        model = model.to(device=self.device)
        criterion = nn.CrossEntropyLoss()
        criterion = criterion.to(device=self.device)
        criterion_smooth = CrossEntropyLabelSmooth(self.num_class,
                                                   self.label_smooth)
        criterion_smooth = criterion_smooth.to(device=self.device)

        ## Preparing Optimizers
        all_parameters = model.parameters()
        weight_parameters = []
        for pname, p in model.named_parameters():
            if (p.ndimension() == 4 or pname == 'classifier.0.weight'
                    or pname == 'classifier.0.bias'):
                weight_parameters.append(p)
        weight_parameters_id = list(map(id, weight_parameters))
        other_parameters = list(
            filter(lambda p: id(p) not in weight_parameters_id,
                   all_parameters))

        optimizer = torch.optim.Adam(
            [
                {
                    'params': other_parameters
                },
                {
                    'params': weight_parameters,
                    'weight_decay': self.weight_decay
                },
            ],
            lr=self.lr,
        )
        ##################

        scheduler = torch.optim.lr_scheduler.LambdaLR(
            optimizer, lambda step: (1.0 - step / self.epochs), last_epoch=-1)

        start_epoch = 0
        best_top1_acc = 0

        os.makedirs(self.save_path, exist_ok=True)
        checkpoint_tar = os.path.join(self.save_path,
                                      f'{self.dataset}-checkpoint.pth.tar')
        if os.path.exists(checkpoint_tar):
            logging.info(
                'loading checkpoint {} ..........'.format(checkpoint_tar))
            checkpoint = torch.load(checkpoint_tar)
            start_epoch = checkpoint['epoch']
            best_top1_acc = checkpoint['best_top1_acc']
            model.load_state_dict(checkpoint['state_dict'], strict=False)
            logging.info('loaded checkpoint {} epoch = {}'.format(
                checkpoint_tar, checkpoint['epoch']))

        # adjust the learning rate according to the checkpoint
        for epoch in range(start_epoch):
            scheduler.step()

        logging.info(
            'epoch, train accuracy, train loss, val accuracy, val loss')
        epoch = start_epoch
        epochs_list = []
        train_top1_acc_list = []
        train_top5_acc_list = []
        val_top1_acc_list = []
        val_top5_acc_list = []
        while epoch < self.epochs:
            train_obj, train_top1_acc, train_top5_acc = self.train(
                epoch,
                self.dataloaders['train'],
                self.model,
                criterion_smooth,
                optimizer,
                scheduler,
            )
            valid_obj, valid_top1_acc, valid_top5_acc = self.validate(
                epoch, self.dataloaders['val'], self.model, criterion,
                self.CFG)

            train_top1_acc_list.append(train_top1_acc)
            train_top5_acc_list.append(train_top5_acc)
            epochs_list.append(epoch)
            val_top1_acc_list.append(valid_top1_acc.cpu().numpy())
            val_top5_acc_list.append(valid_top5_acc.cpu().numpy())
            is_best = False
            if valid_top1_acc > best_top1_acc:
                best_top1_acc = valid_top1_acc
                is_best = True

            save_checkpoint(
                {
                    'epoch': epoch,
                    'state_dict': model.state_dict(),
                    'best_top1_acc': best_top1_acc,
                    'optimizer': optimizer.state_dict(),
                    'scheduler': scheduler.state_dict(),
                },
                is_best,
                self.save_path,
            )

            epoch += 1

        best = torch.load(f'{self.save_path}/model_best.pth.tar')
        self.model.load_state_dict(best['state_dict'])
        self.test((epoch - 1), self.dataloaders['test'], self.model, criterion,
                  self.CFG)
        training_time = (time.time() - start_t) / 36000
        print('total training time = {} hours'.format(training_time))

        df_data = np.array([
            epochs_list,
            train_top1_acc_list,
            train_top5_acc_list,
            val_top1_acc_list,
            val_top5_acc_list,
        ]).T
        df = pd.DataFrame(
            df_data,
            columns=[
                'Epochs',
                'Train Top1',
                'Train Top5',
                'Validation Top1',
                'Validation Top5',
            ],
        )
        df.to_csv(f'{os.getcwd()}/logs/BiRealNet/{self.name}.csv', index=False)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import math
from typing import List
import torch
from torch.nn import BatchNorm2d, Conv2d, Linear
import numpy as np
import pandas as pd

# how to change weights.data https://discuss.pytorch.org/t/how-to-delete-every-grad-after-training/63644/8

__all__ = [
    'update_bn_grad',
    'summary_model',
    'is_depthwise_conv2d',
    'prune_conv2d',
    'prune_bn2d',
    'prune_fc',
    'cal_threshold_by_bn2d_weights',
    'mask2idxes',
    'top_k_idxes',
    'ceil',
    'round_up_to_power_of_2',
]


def update_bn_grad(model, s=0.0001):
    """根据 BN 的 gamma 系数稀疏化训练 BN 层 在 loss.backward() 之后执行.

    Args:
        model: nn.Module
        s: 系数训练的权重

    Returns:
    """
    for m in model.modules():
        if isinstance(m, BatchNorm2d):
            m.weight.grad.data.add_(s * torch.sign(m.weight.data))


def summary_model(model: torch.nn.Module,
                  prune_related_layer_types=(Conv2d, BatchNorm2d, Linear)):
    """打印 model 中和剪枝有关的层."""
    info = []
    for name, module in model.named_modules():
        if type(module) in prune_related_layer_types:
            info.append({'name': name, 'module': module})

    df = pd.DataFrame(info)
    df = df.reindex(columns=['name', 'module'])
    print(df.to_markdown())


def is_depthwise_conv2d(module: torch.nn.Module) -> bool:
    if isinstance(module, torch.nn.Conv2d):
        return module.in_channels == module.out_channels == module.groups
    return False


def prune_bn2d(module: BatchNorm2d, keep_idxes):
    module.num_features = len(keep_idxes)
    module.weight = torch.nn.Parameter(module.weight.data[keep_idxes])
    module.weight.grad = None
    module.bias = torch.nn.Parameter(module.bias.data[keep_idxes])
    module.bias.grad = None
    module.running_mean = module.running_mean[keep_idxes]
    module.running_var = module.running_var[keep_idxes]


def prune_conv2d(module: Conv2d, in_keep_idxes=None, out_keep_idxes=None):
    if in_keep_idxes is None:
        in_keep_idxes = list(range(module.weight.shape[1]))

    if out_keep_idxes is None:
        out_keep_idxes = list(range(module.weight.shape[0]))

    is_depthwise = is_depthwise_conv2d(module)

    if is_depthwise:
        module.groups = len(in_keep_idxes)
        assert len(in_keep_idxes) == len(out_keep_idxes)
    else:
        assert (
            len(in_keep_idxes) <= module.weight.shape[1]
        ), f'len(in_keep_idxes): {len(in_keep_idxes)}, module.weight.shape[1]: {module.weight.shape[1]}'

    assert (
        len(out_keep_idxes) <= module.weight.shape[0]
    ), f'len(out_keep_idxes): {len(out_keep_idxes)}, module.weight.shape[0]: {module.weight.shape[0]}'

    module.out_channels = len(out_keep_idxes)
    module.in_channels = len(in_keep_idxes)

    module.weight = torch.nn.Parameter(
        module.weight.data[out_keep_idxes, :, :, :])

    if not is_depthwise:
        module.weight = torch.nn.Parameter(
            module.weight.data[:, in_keep_idxes, :, :])

    module.weight.grad = None

    if module.bias is not None:
        module.bias = torch.nn.Parameter(module.bias.data[out_keep_idxes])
        module.bias.grad = None

    return in_keep_idxes, out_keep_idxes


def prune_fc(module: Linear,
             keep_idxes: List[int],
             bn_num_channels: int = None):
    """

    Args:
        module:
        keep_idxes:
        bn_num_channels: prev bn num_channels

    Returns:

    """
    if bn_num_channels is not None:
        assert module.in_features % bn_num_channels == 0

        channel_step = module.in_features // bn_num_channels

        _keep_idxes = []
        for idx in keep_idxes:
            _keep_idxes.extend(
                np.asarray(list(range(channel_step))) + idx * channel_step)

        keep_idxes = _keep_idxes

    module.in_features = len(keep_idxes)
    module.weight = torch.nn.Parameter(module.weight.data[:, keep_idxes])
    module.weight.grad = None
    return keep_idxes


def cal_threshold_by_bn2d_weights(bn2d_list: List[BatchNorm2d],
                                  sparsity: float):
    """
    sparsity: 要剪枝的比例
    """
    assert 0 < sparsity < 1

    bn_weight_list = []
    for module in bn2d_list:
        bn_weight_list.append(module.weight.data.cpu().abs().clone())

    bn_weights = torch.cat(bn_weight_list)
    k = int(bn_weights.shape[0] * sparsity)

    sorted_bn = torch.sort(bn_weights)[0]
    thresh = sorted_bn[k]
    return thresh


def mask2idxes(mask):
    idxes = np.squeeze(np.argwhere(mask))
    if idxes.size == 1:
        idxes = np.resize(idxes, (1, ))
    return idxes


def top_k_idxes(module, ratio: float = None, k: int = None):
    if k is not None:
        assert k > 0
    assert ratio is not None or k is not None
    assert not (ratio is None and k is None)

    weights = module.weight.data.abs().clone()

    if ratio is not None:
        k = max(int(weights.shape[0] * ratio), 2)

    idxes = torch.topk(weights.view(-1), k, largest=True)[1]
    return idxes.cpu().numpy()


def ceil(num: int, val: int) -> int:
    # c++ version: num + (val-1) & ~(val-1)
    return int(math.ceil(num / val) * val)


def round_up_to_power_of_2(num: int) -> int:
    # leetcode 231
    n = num - 1
    n |= n >> 1
    n |= n >> 2
    n |= n >> 4
    n |= n >> 8
    n |= n >> 16
    if n < 0:
        return 1
    else:
        return n + 1



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import torch.nn.functional as F
from trailmet.utils.functions import pdist

__all__ = ['KDTransferLoss', 'RkdDistance', 'RKdAngle']


class KDTransferLoss(nn.Module):

    def __init__(self, temperature, reduction):
        super().__init__()
        self.temperature = temperature
        self.reduction = reduction
        self.kd_fun = nn.KLDivLoss(reduction=self.reduction)

    def forward(self, out_t, out_s):
        s_max = F.log_softmax(out_s / self.temperature, dim=1)
        t_max = F.softmax(out_t / self.temperature, dim=1)
        loss_kd = self.kd_fun(s_max, t_max)
        return loss_kd


class RkdDistance(nn.Module):

    def forward(self, student, teacher):
        with torch.no_grad():
            t_d = pdist(teacher, squared=False)
            mean_td = t_d[t_d > 0].mean()
            t_d = t_d / mean_td

        d = pdist(student, squared=False)
        mean_d = d[d > 0].mean()
        d = d / mean_d

        loss = F.smooth_l1_loss(d, t_d, reduction='elementwise_mean')
        return loss


class RKdAngle(nn.Module):

    def forward(self, student, teacher):
        # N x C
        # N x N x C

        with torch.no_grad():
            td = teacher.unsqueeze(0) - teacher.unsqueeze(1)
            norm_td = F.normalize(td, p=2, dim=2)
            t_angle = torch.bmm(norm_td, norm_td.transpose(1, 2)).view(-1)

        sd = student.unsqueeze(0) - student.unsqueeze(1)
        norm_sd = F.normalize(sd, p=2, dim=2)
        s_angle = torch.bmm(norm_sd, norm_sd.transpose(1, 2)).view(-1)

        loss = F.smooth_l1_loss(s_angle, t_angle, reduction='elementwise_mean')
        return loss



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from .utils import adjust_learning_rate
import os
import numpy as np
import pandas as pd
import torch
import torch.optim as optim
import torch.nn as nn
from tqdm import tqdm as tqdm_notebook


class BaseAlgorithm(object):
    """Base Algorithm class that defines the structure of each model
    compression algorithm implemented in this library. Every new algorithm is
    expected to directly use or overwrite the template functions defined below.

    The root command to invoke the compression of any model is
    .compress_model(). Thus, it is required that all algorithms complete this
    template function and use it as the first point of invoking the model
    compression process.

    For methods that require to perform pretraining and fine-tuning, the
    implementation of base_train() method can directly be used for both the
    tasks. In case of modifications, overwrite this function based on the
    needs.
    """

    def __init__(self, **kwargs):
        self.pretraining_epochs = 200
        self.cuda_id = kwargs.get('cuda_id', 0)
        self.device = torch.device(f'cuda:{str(self.cuda_id)}')
        self.log_name = kwargs.get('log_dir', 'abc')
        if os.path.exists('logs') is False:
            os.mkdir('logs')

        if os.path.exists('checkpoints') is False:
            os.mkdir('checkpoints')

    def compress_model(self) -> None:
        """Template function to be overwritten for each model compression
        method."""
        pass

    def base_train(self, model, dataloaders, **kwargs) -> None:
        """This function is used to perform standard model training and can be
        used for various purposes, such as model pretraining, fine-tuning of
        compressed models, among others.

        For cases, where base_train is not directly applicable, feel free to
        overwrite wherever this parent class is inherited.
        """
        best_acc = 0  # setting to lowest possible value
        num_epochs = kwargs.get('EPOCHS', self.pretraining_epochs)
        test_only = kwargs.get('TEST_ONLY', False)

        ### preparing optimizer ###
        optimizer_name = kwargs.get('OPTIMIZER', 'SGD')
        lr = kwargs.get('LR', 0.05)
        scheduler_type = kwargs.get('SCHEDULER_TYPE', 1)
        weight_decay = kwargs.get('WEIGHT_DECAY', 0.001)

        optimizer = self.get_optimizer(optimizer_name, model, lr, weight_decay)
        ###########################

        criterion = nn.CrossEntropyLoss()
        train_losses = []
        valid_losses = []
        valid_accuracy = []

        if test_only is False:
            for epoch in range(num_epochs):
                adjust_learning_rate(optimizer, epoch, num_epochs,
                                     scheduler_type, lr)
                t_loss = self.train_one_epoch(model, dataloaders['train'],
                                              criterion, optimizer)
                acc, v_loss = self.test(model, dataloaders['val'], criterion)
                if acc > best_acc:
                    print('**Saving model**')
                    best_acc = acc
                    torch.save(
                        {
                            'epoch': epoch + 1,
                            'state_dict': model.state_dict(),
                            'acc': best_acc,
                        },
                        f'checkpoints/{self.log_name}.pth',
                    )
                train_losses.append(t_loss)
                valid_losses.append(v_loss)
                valid_accuracy.append(acc)
                df_data = np.array(
                    [train_losses, valid_losses, valid_accuracy]).T
                df = pd.DataFrame(
                    df_data,
                    columns=['train_losses', 'valid_losses', 'valid_accuracy'])
                df.to_csv(f'logs/{self.log_name}.csv')

        state = torch.load(f'checkpoints/{self.log_name}.pth')
        model.load_state_dict(state['state_dict'], strict=True)
        acc, v_loss = self.test(model, dataloaders['test'], criterion)
        print(f"Test Accuracy: {acc} | Valid Accuracy: {state['acc']}")

    def get_optimizer(self, optimizer_name: str, model, lr, weight_decay):
        """Returns the optimizer with the given name."""
        if optimizer_name == 'SGD':
            optimizer = optim.SGD(model.parameters(),
                                  lr=lr,
                                  weight_decay=weight_decay)
        else:
            raise ValueError('Unknown optimizer: %s' % optimizer_name)
        return optimizer

    def train_one_epoch(self,
                        model,
                        dataloader,
                        loss_fn,
                        optimizer,
                        extra_functionality=None):
        """Standard training loop which can be used for various purposes with
        an extra functionality function to add to its working at the end of the
        loop."""
        model.train()
        counter = 0
        tk1 = tqdm_notebook(dataloader, total=len(dataloader))
        running_loss = 0
        for x_var, y_var in tk1:
            counter += 1
            x_var = x_var.to(device=self.device)
            y_var = y_var.to(device=self.device)
            scores = model(x_var)

            loss = loss_fn(scores, y_var)
            running_loss += loss.item()
            tk1.set_postfix(loss=running_loss / counter)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            if extra_functionality is not None:
                extra_functionality()
        return running_loss / counter

    def accuracy(self, output, target, topk=(1, )):
        """Computes the accuracy over the k top predictions for the specified
        values of k."""
        with torch.no_grad():
            maxk = max(topk)
            batch_size = target.size(0)

            _, pred = output.topk(maxk, 1, True, True)
            pred = pred.t()
            correct = pred.eq(target.view(1, -1).expand_as(pred))

            res = []
            for k in topk:
                correct_k = correct[:k].reshape(-1).float().sum(0,
                                                                keepdim=True)
                res.append(correct_k.mul_(100.0 / batch_size))
            return res

    def test(self, model, dataloader, loss_fn=None, device=None):
        """This method is used to test the performance of the trained model."""
        if device is None:
            device = next(model.parameters()).device
        else:
            model.to(device)
        model.eval()
        counter = 0
        tk1 = tqdm_notebook(dataloader, total=len(dataloader))
        running_acc1 = 0
        running_acc5 = 0
        running_loss = 0
        with torch.no_grad():
            for images, targets in tk1:
                counter += 1
                images = images.to(device)
                targets = targets.to(device)
                outputs = model(images)
                acc1, acc5 = self.accuracy(outputs, targets, topk=(1, 5))
                running_acc1 += acc1[0].item()
                running_acc5 += acc5[0].item()
                if loss_fn is not None:
                    loss = loss_fn(outputs, targets)
                    running_loss += loss.item()
                    tk1.set_postfix(
                        loss=running_loss / counter,
                        acc1=running_acc1 / counter,
                        acc5=running_acc5 / counter,
                    )
                else:
                    tk1.set_postfix(acc1=running_acc1 / counter,
                                    acc5=running_acc5 / counter)
        if loss_fn is not None:
            return running_acc1 / counter, running_loss / counter
        return running_acc1 / counter, running_acc5 / counter



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.hub import load_state_dict_from_url
from .base_model import BaseModel

from collections import defaultdict


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding."""
    return nn.Conv2d(in_planes,
                     out_planes,
                     kernel_size=3,
                     stride=stride,
                     padding=1,
                     bias=False)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution."""
    return nn.Conv2d(in_planes,
                     out_planes,
                     kernel_size=1,
                     stride=stride,
                     bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.active = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.active(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.active(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes,
                               planes,
                               kernel_size=3,
                               stride=stride,
                               padding=1,
                               bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes,
                               planes * self.expansion,
                               kernel_size=1,
                               bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion)
        self.active = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.active(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.active(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.active(out)

        return out


class ResNetCifar(BaseModel):

    def __init__(self, block, layers, width=1, num_classes=1000, insize=32):
        super(ResNetCifar, self).__init__()
        self.inplanes = 16
        self.insize = insize
        self.layers_size = layers
        self.num_classes = num_classes
        self.width = width
        self.conv1 = nn.Conv2d(3,
                               16,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.prev_module = defaultdict()
        self.prev_module[self.bn1] = None
        self.active = nn.ReLU(inplace=True)
        self.layer1 = self._make_layer(block, 16 * width, layers[0])
        self.layer2 = self._make_layer(block, 32 * width, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 64 * width, layers[2], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=1)
        self.fc = nn.Linear(64 * width, num_classes)
        self.init_weights()

        assert block is BasicBlock
        prev = self.bn1
        for l_block in [self.layer1, self.layer2, self.layer3]:
            for b in l_block:
                self.prev_module[b.bn1] = prev
                self.prev_module[b.bn2] = b.bn1
                if b.downsample is not None:
                    self.prev_module[b.downsample[1]] = prev
                prev = b.bn2

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            conv_module = nn.Conv2d(
                self.inplanes,
                planes * block.expansion,
                kernel_size=1,
                stride=stride,
                bias=False,
            )
            bn_module = nn.BatchNorm2d(planes * block.expansion)
            if hasattr(bn_module, 'is_imp'):
                bn_module.is_imp = True
            downsample = nn.Sequential(conv_module, bn_module)

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.active(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x

    def get_bn_layers(self):
        bn_layers = []
        for l_blocks in [self.layer1, self.layer2, self.layer3]:
            for b in l_blocks:
                m1, m2 = b.bn1, b.bn2
                bn_layers.append([m1, m2])
        return bn_layers


class ResNet(BaseModel):

    def __init__(
        self,
        block,
        layers,
        width=1,
        num_classes=1000,
        produce_vectors=False,
        init_weights=True,
        insize=32,
    ):
        super(ResNet, self).__init__()
        self.layers_size = layers
        self.num_classes = num_classes
        self.insize = insize
        self.produce_vectors = produce_vectors
        self.block_type = block.__class__.__name__
        self.inplanes = 64
        if insize < 128:
            self.conv1 = nn.Conv2d(3,
                                   64,
                                   kernel_size=3,
                                   stride=1,
                                   padding=1,
                                   bias=False)
        else:
            self.conv1 = nn.Conv2d(3,
                                   64,
                                   kernel_size=7,
                                   stride=2,
                                   padding=3,
                                   bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.prev_module = defaultdict()
        self.prev_module[self.bn1] = None
        self.active = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64 * width, layers[0])
        self.layer2 = self._make_layer(block, 128 * width, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256 * width, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512 * width, layers[3], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=1)  # Global Avg Pool
        self.fc = nn.Linear(512 * block.expansion * width, num_classes)
        self.init_weights()

        for l in [self.layer1, self.layer2, self.layer3, self.layer4]:
            for b in l.children():
                downs = (next(b.downsample.children())
                         if b.downsample is not None else None)

        prev = self.bn1
        for l_block in [self.layer1, self.layer2, self.layer3, self.layer4]:
            for b in l_block:
                self.prev_module[b.bn1] = prev
                self.prev_module[b.bn2] = b.bn1
                self.prev_module[b.bn3] = b.bn2
                if b.downsample is not None:
                    self.prev_module[b.downsample[1]] = prev
                prev = b.bn3

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            conv_module = nn.Conv2d(
                self.inplanes,
                planes * block.expansion,
                kernel_size=1,
                stride=stride,
                bias=False,
            )
            bn_module = nn.BatchNorm2d(planes * block.expansion)
            if hasattr(bn_module, 'is_imp'):
                bn_module.is_imp = True
            downsample = nn.Sequential(conv_module, bn_module)

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.active(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        feature_vectors = x.view(x.size(0), -1)
        x = self.fc(feature_vectors)

        if self.produce_vectors:
            return x, feature_vectors
        else:
            return x

    def get_bn_layers(self):
        bn_layers = []
        for l_blocks in [self.layer1, self.layer2, self.layer3, self.layer4]:
            for b in l_blocks:
                if self.block_type == 'Bottleneck':
                    m1, m2, m3 = b.bn1, b.bn2, b.bn3
                    bn_layers.append([m1, m2, m3])
                else:
                    m1, m2 = b.bn1, b.bn2
                    bn_layers.append([m1, m2])
        return bn_layers


def make_wide_resnet(num_classes, insize):
    model = ResNetCifar(BasicBlock, [4, 4, 4],
                        width=12,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet20(num_classes, insize):
    model = ResNetCifar(BasicBlock, [3, 3, 3],
                        width=1,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet32(num_classes, insize):
    model = ResNetCifar(BasicBlock, [5, 5, 5],
                        width=1,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet50(num_classes, insize):
    model = ResNet(Bottleneck, [3, 4, 6, 3],
                   num_classes=num_classes,
                   insize=insize)
    return model


def make_resnet56(num_classes, insize):
    model = ResNetCifar(BasicBlock, [9, 9, 9],
                        width=1,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet18(num_classes, insize):
    model = ResNet(BasicBlock, [2, 2, 2, 2],
                   num_classes=num_classes,
                   insize=insize)
    return model


def make_resnet101(num_classes, insize):
    model = ResNet(Bottleneck, [3, 4, 23, 3],
                   num_classes=num_classes,
                   insize=insize)
    return model


def make_resnet110(num_classes, insize):
    model = ResNetCifar(BasicBlock, [18, 18, 18],
                        width=1,
                        num_classes=num_classes,
                        insize=insize)
    return model


def make_resnet152(num_classes, insize):
    model = ResNet(Bottleneck, [3, 8, 36, 3],
                   num_classes=num_classes,
                   insize=insize)
    return model


def get_resnet_model(model, num_classes, insize, pretrained):
    """Returns the requested model, ready for training/pruning with the
    specified method.

    :param model: str, either wrn or r50
    :param num_classes: int, num classes in the dataset
    :return: A prunable ResNet model
    """
    if model == 'wrn':
        net = make_wide_resnet(num_classes, insize)
        pretrained_weights = None

    elif model == 'resnet18':
        net = make_resnet18(num_classes, insize)
        pretrained_weights = 'https://download.pytorch.org/models/resnet18-f37072fd.pth'
    elif model == 'resnet20':
        net = make_resnet20(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet32':
        net = make_resnet32(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet50':
        net = make_resnet50(num_classes, insize)
        pretrained_weights = 'https://download.pytorch.org/models/resnet50-11ad3fa6.pth'
    elif model == 'resnet56':
        net = make_resnet56(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet101':
        net = make_resnet101(num_classes, insize)
        pretrained_weights = (
            'https://download.pytorch.org/models/resnet101-cd907fc2.pth')
    elif model == 'resnet110':
        net = make_resnet110(num_classes, insize)
        pretrained_weights = None
    elif model == 'resnet152':
        net = make_resnet152(num_classes, insize)
        pretrained_weights = (
            'https://download.pytorch.org/models/resnet152-f82ba261.pth')
    if pretrained:
        if pretrained_weights != None:
            weights = load_state_dict_from_url(pretrained_weights,
                                               progress=True)
            net.load_state_dict(weights, strict=False)
        else:
            print('pretrained weights not available')
    return net


class BinaryActivation(nn.Module):

    def __init__(self):
        super(BinaryActivation, self).__init__()

    def forward(self, x):
        out_forward = torch.sign(x)
        # out_e1 = (x^2 + 2*x)
        # out_e2 = (-x^2 + 2*x)
        out_e_total = 0
        mask1 = x < -1
        mask2 = x < 0
        mask3 = x < 1
        out1 = (-1) * mask1.type(
            torch.float32) + (x * x + 2 * x) * (1 - mask1.type(torch.float32))
        out2 = out1 * mask2.type(
            torch.float32) + (-x * x + 2 * x) * (1 - mask2.type(torch.float32))
        out3 = out2 * mask3.type(
            torch.float32) + 1 * (1 - mask3.type(torch.float32))
        out = out_forward.detach() - out3.detach() + out3

        return out


class HardBinaryConv(nn.Module):

    def __init__(self, in_chn, out_chn, kernel_size=3, stride=1, padding=1):
        super(HardBinaryConv, self).__init__()
        self.stride = stride
        self.padding = padding
        self.number_of_weights = in_chn * out_chn * kernel_size * kernel_size
        self.shape = (out_chn, in_chn, kernel_size, kernel_size)
        self.weights = nn.Parameter(torch.rand(
            (self.number_of_weights, 1)) * 0.001,
                                    requires_grad=True)

    def forward(self, x):
        real_weights = self.weights.view(self.shape)
        scaling_factor = torch.mean(
            torch.mean(
                torch.mean(torch.abs(real_weights), dim=3, keepdim=True),
                dim=2,
                keepdim=True,
            ),
            dim=1,
            keepdim=True,
        )
        # print(scaling_factor, flush=True)
        scaling_factor = scaling_factor.detach()
        binary_weights_no_grad = scaling_factor * torch.sign(real_weights)
        cliped_weights = torch.clamp(real_weights, -1.0, 1.0)
        binary_weights = (binary_weights_no_grad.detach() -
                          cliped_weights.detach() + cliped_weights)
        #         print(binary_weights.shape)
        y = F.conv2d(x,
                     binary_weights,
                     stride=self.stride,
                     padding=self.padding)

        return y


class BinaryBasicBlock(nn.Module):
    expansion = 1

    def __init__(self,
                 inplanes,
                 planes,
                 stride=1,
                 downsample=None,
                 binarise=True):
        super(BinaryBasicBlock, self).__init__()
        self.binarise = binarise
        if self.binarise:
            self.activation = BinaryActivation()
            self.conv = HardBinaryConv(inplanes, planes, stride=stride)
        else:
            self.activation = nn.ReLU(inplace=True)
            self.conv = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.bn2 = None
        self.bn3 = None
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x
        out = self.activation(x)
        out = self.conv(out)
        out = self.bn1(out)
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual

        return out


class BirealNet(ResNet):

    def __init__(
        self,
        block,
        layers,
        num_fp=0,
        width=1,
        num_classes=1000,
        produce_vectors=False,
        init_weights=True,
        insize=32,
    ):
        super(BirealNet, self).__init__(
            block,
            layers,
            width=1,
            num_classes=num_classes,
            produce_vectors=False,
            init_weights=True,
            insize=32,
        )
        del self.prev_module
        self.num_fp = num_fp
        self.count_fp = 0
        self.inplanes = 64
        self.layer1 = self._make_layer1(block, 64 * width, layers[0])
        self.layer2 = self._make_layer1(block,
                                        128 * width,
                                        layers[1],
                                        stride=2)
        self.layer3 = self._make_layer1(block,
                                        256 * width,
                                        layers[2],
                                        stride=2)
        self.layer4 = self._make_layer1(block,
                                        512 * width,
                                        layers[3],
                                        stride=2)

    def _make_layer1(self, block, planes, blocks, stride=1):
        downsample = None
        if self.count_fp < self.num_fp:
            if stride != 1 or self.inplanes != planes * block.expansion:
                downsample = nn.Sequential(
                    nn.AvgPool2d(kernel_size=2, stride=stride),
                    conv1x1(self.inplanes, planes * block.expansion),
                    nn.BatchNorm2d(planes * block.expansion),
                )

            layers = []
            layers.append(
                block(self.inplanes,
                      planes,
                      stride,
                      downsample,
                      binarise=False))
            self.inplanes = planes * block.expansion
            for _ in range(1, blocks):
                layers.append(block(self.inplanes, planes, binarise=False))
            self.count_fp += 1
            return nn.Sequential(*layers)
        else:
            if stride != 1 or self.inplanes != planes * block.expansion:
                downsample = nn.Sequential(
                    nn.AvgPool2d(kernel_size=2, stride=stride),
                    conv1x1(self.inplanes, planes * block.expansion),
                    nn.BatchNorm2d(planes * block.expansion),
                )

            layers = []
            layers.append(
                block(self.inplanes, planes, stride, downsample,
                      binarise=True))
            self.inplanes = planes * block.expansion
            for _ in range(1, blocks):
                layers.append(block(self.inplanes, planes, binarise=True))
            return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x


def make_birealnet18(num_classes, insize, num_fp=0):
    model = BirealNet(
        BinaryBasicBlock,
        [4, 4, 4, 4],
        num_classes=num_classes,
        insize=insize,
        num_fp=num_fp,
    )
    return model


def make_birealnet34(num_classes, insize, num_fp=0):
    model = BirealNet(
        BinaryBasicBlock,
        [6, 8, 12, 6],
        num_classes=num_classes,
        insize=insize,
        num_fp=num_fp,
    )
    return model


def make_birealnet50(num_classes, insize, num_fp=0):
    model = BirealNet(
        BinaryBasicBlock,
        [3, 4, 6, 3],
        num_classes=num_classes,
        insize=insize,
        num_fp=num_fp,
    )
    return model



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import torch.nn as nn
import torch.utils.data
import torchvision.models as models

from trailmet.algorithms.prune.pruner import pruner_dict
from trailmet.algorithms.prune.prune import BasePruning

import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import numpy as np
import os
import time

from trailmet.utils import (
    AverageMeter,
    accuracy,
    save_checkpoint,
    strlist_to_list,
    adjust_learning_rate,
)

logger = logging.getLogger(__name__)


def is_single_branch(name):
    return False


class Growth_Regularisation(BasePruning):
    """Base Algorithm class that defines the structure of each model
    compression algorithm implemented in this library.

    Every new algorithm is expected to directly use or overwrite the template
    functions defined below. The root command to invoke the compression of any
    model is .compress_model(). Thus, it is required that all algorithms
    complete this template function and use it as the first point of invoking
    the model compression process. For methods that require to perform
    pretraining and fine-tuning, the implementation of base_train() method can
    directly be used for both the tasks. In case of modifications, overwrite
    this function based on the needs.
    """

    def __init__(self, model, dataloaders, **kwargs):
        self.device = 'cuda'
        self.train_loader = dataloaders['train']
        self.val_loader = dataloaders['val']
        self.test_loader = dataloaders['test']
        self.dataloaders = dataloaders
        self.model = model
        self.kwargs = kwargs
        self.dataset = self.kwargs.get('DATASET', 'c100')
        self.num_classes = self.kwargs.get('num_classes', 100)
        self.label_smooth = self.kwargs.get('label_smooth', 0.1)
        self.pretrained = self.kwargs.get('pretrained', '')
        self.resume = self.kwargs.get('resume', 'False')
        self.epochs = self.kwargs.get('epoch', '120')
        self.weight_decay = self.kwargs.get('weight_decay', 0)
        self.learning_rate = self.kwargs.get('learning_rate', 0.001)
        self.stage_pr = self.kwargs.get('stage_pr', None)
        self.skip_layers = self.kwargs.get('skip_layers', None)
        self.base_pr_model = self.kwargs.get('base_pr_model', None)
        self.base_model_path = self.kwargs.get('base_model_path', None)
        self.momentum = self.kwargs.get('momentum', 0.9)
        self.save = self.kwargs.get('save', './checkpoint')
        self.method = self.kwargs.get('method', 'GReg-1')

        self.wandb_monitor = self.kwargs.get('wandb', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__

        self.name = '_'.join([
            self.dataset_name,
            f'{self.epochs}',
            f'{self.learning_rate}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/Growth_Regularisation', exist_ok=True)
        os.makedirs(self.save, exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/Growth_Regularisation/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.kwargs}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet Growth_Regularisation',
                       name=self.name)
            wandb.config.update(self.kwargs)

        if self.stage_pr:
            self.stage_pr = self.stage_pr  # example: [0, 0.4, 0.5, 0]
            self.skip_layers = strlist_to_list(self.skip_layers,
                                               str)  # example: [2.3.1, 3.1]
        else:
            assert (
                self.base_pr_model
            ), 'If stage_pr is not provided, base_pr_model must be provided'

    def compress_model(self) -> None:
        """Template function to be overwritten for each model compression
        method."""

        self.epochs = 1

        self.model.maxpool = torch.nn.Identity()
        self.model = self.base_train(self.model,
                                     self.dataloaders,
                                     fine_tune=False)

        self.prune_and_finetune(self.kwargs, self.dataloaders)

    def prune_and_finetune(self, args, dataloader):
        """Prune and finetune the model."""

        # if self.base_model_path != None:
        if 0:
            X = torch.load(self.base_model_path)
            X1 = X['state_dict']
            L = list(X1.keys())
            for key in L:
                new_key = key.replace('model.', '')
                X1[new_key] = X1.pop(key)
            self.model.load_state_dict(X1)
            print("==> Load pretrained model successfully: '%s'" %
                  self.base_model_path)
            logger.info("==> Load pretrained model successfully: '%s'" %
                        self.base_model_path)

        criterion = nn.CrossEntropyLoss()

        optimizer = torch.optim.SGD(
            self.model.parameters(),
            self.learning_rate,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
        )
        prune_state, pruner = '', None
        if prune_state != 'finetune':

            class passer:
                pass  # to pass arguments

            passer.train_loader = dataloader['train']
            passer.test_loader = dataloader['val']
            passer.save = save_checkpoint
            passer.criterion = criterion
            passer.train_sampler = None
            passer.pruner = pruner
            passer.args = self.kwargs
            passer.is_single_branch = is_single_branch
            pruner = pruner_dict[self.method](self.model, self.kwargs, logger,
                                              passer)
            if self.method == 'L1':
                model = pruner.prune()
                print('==> Saving model without key <==')
                print(model)
                save_checkpoint(
                    {
                        'arch': args.arch,
                        'model': model,
                        'state_dict': model.state_dict(),
                    },
                    False,
                    self.save,
                )
            else:
                pruning_key, model = pruner.prune()  # get the pruned model
                print('==> Saving model with key <==')
                print(model)
                save_checkpoint(
                    {
                        'arch': args.arch,
                        'model': model,
                        'state_dict': model.state_dict(),
                        'pruning_key': pruning_key,
                    },
                    False,
                    self.save,
                )
        self.base_train(model, dataloader, pruning_key)

    def base_train(self, model, dataloaders, fine_tune=False):
        """This function is used to perform standard model training and can be
        used for various purposes, such as model pretraining, fine-tuning of
        compressed models, among others.

        For cases, where base_train is not directly applicable, feel free to
        overwrite wherever this parent class is inherited.
        """
        best_top1_acc = 0  # setting to lowest possible value
        scheduler_type = 1
        self.pr = None
        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=self.learning_rate,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
        )
        ###########################

        self.model.to(self.device)
        criterion = nn.CrossEntropyLoss()
        epochs_list = []
        val_top1_acc_list = []
        val_top5_acc_list = []

        for epoch in range(self.epochs):
            adjust_learning_rate(optimizer, epoch, self.epochs, scheduler_type,
                                 self.learning_rate)

            t_loss = self.train_one_epoch(
                model,
                dataloaders['train'],
                criterion,
                optimizer,
                epoch,
            )

            valid_loss, valid_top1_acc, valid_top5_acc = self.test(
                model,
                dataloaders['val'],
                criterion,
                epoch,
            )

            is_best = False
            if valid_top1_acc > best_top1_acc:
                best_top1_acc = valid_top1_acc
                is_best = True

                if self.pr is not None:
                    print('==> Saving model with key <==')
                    logger.info('==> Saving model with key <==')
                    save_checkpoint(
                        {
                            'epoch': epoch,
                            'state_dict': model.state_dict(),
                            'best_top1_acc': best_top1_acc,
                            'pruning_key': self.pr,
                            'optimizer': optimizer.state_dict(),
                        },
                        is_best,
                        self.save,
                    )

                else:
                    print('==> Saving model without key <==')
                    logger.info('==> Saving model without key <==')
                    save_checkpoint(
                        {
                            'epoch': epoch,
                            'state_dict': model.state_dict(),
                            'best_top1_acc': best_top1_acc,
                            'pruning_key': self.pr,
                            'optimizer': optimizer.state_dict(),
                        },
                        is_best,
                        self.save,
                    )

                if self.wandb_monitor:
                    wandb.log({'best_top1_acc': best_top1_acc})

                epochs_list.append(epoch)
                val_top1_acc_list.append(valid_top1_acc.cpu().numpy())
                val_top5_acc_list.append(valid_top5_acc.cpu().numpy())

                df_data = np.array([
                    epochs_list,
                    val_top1_acc_list,
                    val_top5_acc_list,
                ]).T
                df = pd.DataFrame(
                    df_data,
                    columns=[
                        'Epochs',
                        'Validation Top1',
                        'Validation Top5',
                    ],
                )
                df.to_csv(
                    f'{os.getcwd()}/logs/Growth_Regularisation/{self.name}.csv',
                    index=False,
                )

        return model

    def train_one_epoch(
        self,
        model,
        dataloader,
        loss_fn,
        optimizer,
        epoch,
        extra_functionality=None,
    ):
        """Standard training loop which can be used for various purposes with
        an extra functionality function to add to its working at the end of the
        loop."""
        model.train()

        batch_time = AverageMeter('Time', ':6.3f')
        data_time = AverageMeter('Data', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')

        end = time.time()

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Training Epoch [X] (X / X Steps) (batch time=X.Xs) (data time=X.Xs) (loss=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        for i, (images, labels) in enumerate(epoch_iterator):
            data_time.update(time.time() - end)
            images = images.to(device=self.device)
            labels = labels.to(device=self.device)
            scores = model(images)

            loss = loss_fn(scores, labels)
            n = images.size(0)
            losses.update(loss.item(), n)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            batch_time.update(time.time() - end)
            end = time.time()

            epoch_iterator.set_description(
                'Training Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(dataloader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                ))

            logger.info(
                'Training Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(dataloader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                ))

            if self.wandb_monitor:
                wandb.log({
                    f'train_loss': losses.val,
                })

            if extra_functionality is not None:
                extra_functionality()

        return losses.avg

    def test(self, model, dataloader, loss_fn=None, epoch=0):
        """This method is used to test the performance of the trained model."""

        model.eval()
        model.to(self.device)

        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Validating Epoch [X] (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        with torch.no_grad():
            end = time.time()
            for i, (images, targets) in enumerate(epoch_iterator):
                images = images.to(self.device)
                targets = targets.to(self.device)
                outputs = model(images)
                pred1, pred5 = accuracy(outputs, targets, topk=(1, 5))

                if loss_fn is not None:
                    loss = loss_fn(outputs, targets)

                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        f'val_loss': losses.val,
                        f'val_top1_acc': top1.val,
                        f'val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))

        return losses.avg, top1.avg, top5.avg



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import copy
import torch
import torch.nn as nn
import torch.distributed as dist
from trailmet.utils import seed_everything
from trailmet.algorithms.quantize.quantize import (
    BaseQuantization,
    FoldBN,
    StraightThrough,
)
from trailmet.models.resnet import BasicBlock, Bottleneck
from trailmet.models.mobilenet import InvertedResidual
from trailmet.algorithms.quantize.qmodel import (
    QuantBasicBlock,
    QuantBottleneck,
    QuantInvertedResidual,
    QuantModule,
    BaseQuantBlock,
)
from trailmet.algorithms.quantize.reconstruct import (
    layer_reconstruction,
    block_reconstruction,
)

import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import numpy as np
import os
import time

from trailmet.utils import AverageMeter, accuracy, save_checkpoint

logger = logging.getLogger(__name__)

supported = {
    BasicBlock: QuantBasicBlock,
    Bottleneck: QuantBottleneck,
    InvertedResidual: QuantInvertedResidual,
}


class BRECQ(BaseQuantization):
    """
    Class for post-training quantization using block reconstruction method
    based on - BRECQ: PUSHING THE LIMIT OF POST-TRAINING QUANTIZATION
    BY BLOCK RECONSTRUCTION [https://arxiv.org/abs/2102.05426]

    :param W_BITS: bitwidth for weight quantization
    :param A_BITS: bitwidth for activation quantization
    :param CHANNEL_WISE: apply channel_wise quantization for weights
    :param ACT_QUANT: apply activation quantization
    :param SET_8BIT_HEAD_STEM: Set the first and the last layer to 8-bit
    :param NUM_SAMPLES: size of calibration dataset
    :param WEIGHT: weight of rounding cost vs the reconstruction loss
    :param ITERS_W: number of iteration for AdaRound
    :param ITERS_A: number of iteration for LSQ
    :param LR: learning rate for LSQ
    """

    def __init__(self, model: nn.Module, dataloaders, **kwargs):
        super(BRECQ, self).__init__(**kwargs)
        self.model = copy.deepcopy(model)
        self.train_loader = dataloaders['train']
        self.test_loader = dataloaders['test']
        self.kwargs = kwargs
        self.w_bits = self.kwargs.get('W_BITS', 8)
        self.a_bits = self.kwargs.get('A_BITS', 8)
        self.channel_wise = self.kwargs.get('CHANNEL_WISE', True)
        self.act_quant = self.kwargs.get('ACT_QUANT', True)
        self.set_8bit_head_stem = self.kwargs.get('SET_8BIT_HEAD_STEM', False)
        self.precision_config = self.kwargs.get('PREC_CONFIG', [])
        self.num_samples = self.kwargs.get('NUM_SAMPLES', 1024)
        self.weight = self.kwargs.get('WEIGHT', 0.01)
        self.iters_w = self.kwargs.get('ITERS_W', 10000)
        self.iters_a = self.kwargs.get('ITERS_A', 10000)
        self.optimizer = self.kwargs.get('OPTIMIZER', 'adam')
        self.lr = self.kwargs.get('LR', 4e-4)
        self.gpu_id = self.kwargs.get('GPU_ID', 0)
        self.calib_bs = self.kwargs.get('CALIB_BS', 64)
        self.seed = self.kwargs.get('SEED', 42)
        self.p = 2.4  # Lp norm minimization for LSQ
        self.b_start = 20  # temperature at the beginning of calibration
        self.b_end = 2  # temperature at the end of calibration
        self.test_before_calibration = True
        self.device = torch.device('cuda:{}'.format(self.gpu_id))
        torch.cuda.set_device(self.gpu_id)
        seed_everything(self.seed)
        print('==> Using seed :', self.seed)

        self.wandb_monitor = self.kwargs.get('WANDB', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__
        self.save = './checkpoints/'

        self.name = '_'.join([
            self.dataset_name,
            f'{self.a_bits}',
            f'{self.lr}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/BRECQ', exist_ok=True)
        os.makedirs(self.save, exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/BRECQ/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.kwargs}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet BRECQ', name=self.name)
            wandb.config.update(self.kwargs)

    def compress_model(self):
        """Method to build quantization parameters and finetune weights and/or
        activations."""
        wq_params = {
            'n_bits': self.w_bits,
            'channel_wise': self.channel_wise,
            'scale_method': 'mse',
        }
        aq_params = {
            'n_bits': self.a_bits,
            'channel_wise': False,
            'scale_method': 'mse',
            'leaf_param': self.act_quant,
        }
        self.model = self.model.to(self.device)
        self.model.eval()
        self.qnn = QuantModel(model=self.model,
                              weight_quant_params=wq_params,
                              act_quant_params=aq_params)
        self.qnn = self.qnn.to(self.device)
        self.qnn.eval()

        for i in range(len(self.precision_config)):
            conf = self.precision_config[i]
            self.qnn.set_layer_precision(conf[2], conf[3], conf[0], conf[1])
            print(
                f'==> Layers from {conf[0]} to {conf[1]} set to precision w{conf[2]}a{conf[3]}'
            )
            logger.info(
                f'==> Layers from {conf[0]} to {conf[1]} set to precision w{conf[2]}a{conf[3]}'
            )

        if self.set_8bit_head_stem:
            print('==> Setting the first and the last layer to 8-bit')
            logger.info('==> Setting the first and the last layer to 8-bit')
            self.qnn.set_first_last_layer_to_8bit()

        self.cali_data = self.get_calib_samples(self.train_loader,
                                                self.num_samples)
        # device = next(self.qnn.parameters()).device

        # Initialize weight quantization parameters
        self.qnn.set_quant_state(True, False)
        print('==> Initializing weight quantization parameters')
        logger.info('==> Initializing weight quantization parameters')
        _ = self.qnn(self.cali_data[:self.calib_bs].to(self.device))
        if self.test_before_calibration:
            valid_loss, valid_top1_acc, valid_top5_acc = self.test(
                self.qnn, self.test_loader, nn.CrossEntropyLoss())
            print('Quantized accuracy before brecq: {}'.format(valid_top1_acc))
            logger.info(
                'Quantized accuracy before brecq: {}'.format(valid_top1_acc))

        # Start weight calibration
        kwargs = dict(
            cali_data=self.cali_data,
            iters=self.iters_w,
            weight=self.weight,
            asym=True,
            b_range=(self.b_start, self.b_end),
            warmup=0.2,
            act_quant=False,
            opt_mode='mse',
            optim=self.optimizer,
        )
        print('==> Starting weight calibration')
        logger.info('==> Starting weight calibration')
        self.reconstruct_model(self.qnn, **kwargs)
        self.qnn.set_quant_state(weight_quant=True, act_quant=False)
        valid_loss, valid_top1_acc, valid_top5_acc = self.test(
            self.qnn, self.test_loader, nn.CrossEntropyLoss())
        print('Weight quantization accuracy: {}'.format(valid_top1_acc))

        if self.act_quant:
            # Initialize activation quantization parameters
            self.qnn.set_quant_state(True, True)
            with torch.no_grad():
                _ = self.qnn(self.cali_data[:self.calib_bs].to(self.device))

            # Disable output quantization because network output
            # does not get involved in further computation
            self.qnn.disable_network_output_quantization()

            # Start activation rounding calibration
            kwargs = dict(
                cali_data=self.cali_data,
                iters=self.iters_a,
                act_quant=True,
                opt_mode='mse',
                lr=self.lr,
                p=self.p,
                optim=self.optimizer,
            )
            self.reconstruct_model(self.qnn, **kwargs)
            self.qnn.set_quant_state(weight_quant=True, act_quant=True)
            valid_loss, valid_top1_acc, valid_top5_acc = self.test(
                self.qnn, self.test_loader, nn.CrossEntropyLoss())
            print('Full quantization (W{}A{}) accuracy: {}'.format(
                self.w_bits, self.a_bits, valid_top1_acc))
            logger.info('Full quantization (W{}A{}) accuracy: {}'.format(
                self.w_bits, self.a_bits, valid_top1_acc))

        return self.qnn

    def reconstruct_model(self, model: nn.Module, **kwargs):
        """Method for model parameters reconstruction.

        Takes in quantized model and optimizes weights by applying layer-wise
        reconstruction for first and last layer, and block reconstruction
        otherwise.
        """
        for name, module in model.named_children():
            if isinstance(module, QuantModule):
                if module.ignore_reconstruction is True:
                    print('Ignore reconstruction of layer {}'.format(name))
                    logger.info(
                        'Ignore reconstruction of layer {}'.format(name))
                    continue
                else:
                    print('Reconstruction for layer {}'.format(name))
                    logger.info('Reconstruction for layer {}'.format(name))
                    layer_reconstruction(self.qnn, module, **kwargs)
            elif isinstance(module, BaseQuantBlock):
                if module.ignore_reconstruction is True:
                    print('Ignore reconstruction of block {}'.format(name))
                    logger.info(
                        'Ignore reconstruction of block {}'.format(name))
                    continue
                else:
                    print('Reconstruction for block {}'.format(name))
                    logger.info('Reconstruction for block {}'.format(name))
                    block_reconstruction(self.qnn, module, **kwargs)
            else:
                self.reconstruct_model(module, **kwargs)

    def test(self, model, dataloader, loss_fn):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Validating network (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        model.eval()
        model.to(self.device)

        with torch.no_grad():
            end = time.time()

            for i, (images, labels) in enumerate(epoch_iterator):
                images = images.to(self.device, dtype=torch.float)
                labels = labels.to(self.device)

                preds = model(images)

                loss = loss_fn(preds, labels)

                pred1, pred5 = accuracy(preds, labels, topk=(1, 5))

                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating network (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating network (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'val_loss': losses.val,
                        'val_top1_acc': top1.val,
                        'val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))
        return losses.avg, top1.avg, top5.avg


class QuantModel(nn.Module):
    """Recursively replace the normal conv2d and Linear layer to QuantModule,
    to enable calculating activation statistics and storing scaling factors.

    :param module: nn.Module with nn.Conv2d or nn.Linear in its children
    :param weight_quant_params: quantization parameters like n_bits for weight
        quantizer
    :param act_quant_params: quantization parameters like n_bits for activation
        quantizer
    """

    def __init__(
        self,
        model: nn.Module,
        weight_quant_params: dict = {},
        act_quant_params: dict = {},
    ):
        super().__init__()
        self.model = model
        bn = FoldBN()
        bn.search_fold_and_remove_bn(self.model)
        self.quant_module_refactor(self.model, weight_quant_params,
                                   act_quant_params)
        self.quant_modules = [
            m for m in self.model.modules() if isinstance(m, QuantModule)
        ]

    def quant_module_refactor(
        self,
        module: nn.Module,
        weight_quant_params: dict = {},
        act_quant_params: dict = {},
    ):
        prev_quantmodule = None
        for name, child_module in module.named_children():
            if type(child_module) in supported:
                setattr(
                    module,
                    name,
                    supported[type(child_module)](child_module,
                                                  weight_quant_params,
                                                  act_quant_params),
                )

            elif isinstance(child_module, (nn.Conv2d, nn.Linear)):
                setattr(
                    module,
                    name,
                    QuantModule(child_module, weight_quant_params,
                                act_quant_params),
                )
                prev_quantmodule = getattr(module, name)

            elif isinstance(child_module, (nn.ReLU, nn.ReLU6)):
                if prev_quantmodule is not None:
                    prev_quantmodule.activation_function = child_module
                    setattr(module, name, StraightThrough())
                else:
                    continue

            elif isinstance(child_module, StraightThrough):
                continue

            else:
                self.quant_module_refactor(child_module, weight_quant_params,
                                           act_quant_params)

    def set_quant_state(self,
                        weight_quant: bool = False,
                        act_quant: bool = False):
        """
        :param weight_quant: set True for weight quantization
        :param act_quant: set True for activation quantization
        """
        for m in self.model.modules():
            if isinstance(m, (QuantModule, BaseQuantBlock)):
                m.set_quant_state(weight_quant, act_quant)

    def quantize_model_till(self, layer, act_quant: bool = False):
        """
        :param layer: block/layer upto which model is to be quantized.
        :param act_quant: set True for activation quantization
        """
        self.set_quant_state(False, False)
        for name, module in self.model.named_modules():
            if isinstance(module, (QuantModule, BaseQuantBlock)):
                module.set_quant_state(True, act_quant)
            if module == layer:
                break

    def forward(self, input):
        return self.model(input)

    def set_first_last_layer_to_8bit(self):
        """Set the precision (bitwidth) used for quantizing weights and
        activations to 8-bit for the first and last layers of the model.

        Also ignore reconstruction for the first layer.
        """
        assert (len(self.quant_modules)
                >= 2), 'Model has less than 2 quantization modules'
        self.quant_modules[0].weight_quantizer.bitwidth_refactor(8)
        self.quant_modules[0].act_quantizer.bitwidth_refactor(8)
        self.quant_modules[-1].weight_quantizer.bitwidth_refactor(8)
        self.quant_modules[-2].act_quantizer.bitwidth_refactor(8)
        self.quant_modules[0].ignore_reconstruction = True

    def disable_network_output_quantization(self):
        self.quant_modules[-1].disable_act_quant = True

    def set_layer_precision(self, weight_bit=8, act_bit=8, start=0, end=None):
        """Set the precision (bitwidth) used for quantizing weights and
        activations for a range of layers in the model.

        :param weight_bit: number of bits to use for quantizing weights
        :param act_bit: number of bits to use for quantizing activations
        :param start: index of the first layer to set the precision for
            (default: 0)
        :param end: index of the last layer to set the precision for (default:
            None, i.e., the last layer)
        """
        assert start >= 0 and end >= 0, 'layer index cannot be negative'
        assert start < len(self.quant_modules) and end < len(
            self.quant_modules), 'layer index out of range'

        for module in self.quant_modules[start:end + 1]:
            module.weight_quantizer.bitwidth_refactor(weight_bit)
            if module is not self.quant_modules[-1]:
                module.act_quantizer.bitwidth_refactor(act_bit)

    def synchorize_activation_statistics(self):
        """Synchronize the statistics of the activation quantizers across all
        distributed workers."""
        for m in self.modules():
            if isinstance(m, QuantModule):
                if m.act_quantizer.delta is not None:
                    m.act_quantizer.delta.data /= dist.get_world_size()
                    dist.all_reduce(m.act_quantizer.delta.data)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F

import logging
from datetime import datetime
from tqdm import tqdm
import wandb
import pandas as pd
import numpy as np
import os
import time

from trailmet.utils import AverageMeter, accuracy, save_checkpoint

logger = logging.getLogger(__name__)

from trailmet.algorithms.distill.distill import Distillation, ForwardHookManager

# Paying More Attention to Attention: Improving the Performance of Convolutional Neural Networks via Attention Transfer


class AttentionTransferLoss(nn.Module):

    def __init__(self):
        super().__init__()

    @staticmethod
    def attention_map(feature_map):
        """Compute the attention map from a feature map."""
        return F.normalize(feature_map.pow(2).mean(1).flatten(1))

    def compute_loss(self, teacher_feature_map, student_feature_map):
        """Compute the loss between teacher and student feature maps."""
        teacher_attention_map = self.attention_map(teacher_feature_map)
        student_attention_map = self.attention_map(student_feature_map)
        loss = (teacher_attention_map - student_attention_map).pow(2).mean()
        return loss

    def forward(self, feature_map_pairs):
        """feature_map_pairs: list of (teacher_feature_map, student_feature_map)"""
        loss = 0
        for teacher_feature_map, student_feature_map in feature_map_pairs:
            loss += self.compute_loss(teacher_feature_map, student_feature_map)
        return loss


class AttentionTransfer(Distillation):
    """Class to compress model using distillation via attention transfer."""

    def __init__(self, teacher_model, student_model, dataloaders, **kwargs):
        super(AttentionTransfer, self).__init__(**kwargs)
        self.teacher_model = teacher_model
        self.student_model = student_model
        self.dataloaders = dataloaders
        self.kwargs = kwargs

        self.device = kwargs['DEVICE']
        self.beta = self.kwargs['DISTILL_ARGS'].get('BETA', 1000)

        # self.student_io_dict, self.teacher_io_dict = dict(), dict()
        self.teacher_layer_names = kwargs['DISTILL_ARGS'].get(
            'TEACHER_LAYER_NAMES')
        self.student_layer_names = kwargs['DISTILL_ARGS'].get(
            'STUDENT_LAYER_NAMES')
        self.forward_hook_manager_teacher = ForwardHookManager(self.device)
        self.forward_hook_manager_student = ForwardHookManager(self.device)

        self.ce_loss = nn.CrossEntropyLoss()
        self.at_loss = AttentionTransferLoss()

        self.epochs = kwargs['DISTILL_ARGS'].get('EPOCHS', 200)
        self.lr = kwargs['DISTILL_ARGS'].get('LR', 0.1)

        self.wandb_monitor = self.kwargs.get('wandb', 'False')
        self.dataset_name = dataloaders['train'].dataset.__class__.__name__
        self.save = './checkpoints/'

        self.name = '_'.join([
            self.dataset_name,
            f'{self.epochs}',
            f'{self.lr}',
            datetime.now().strftime('%b-%d_%H:%M:%S'),
        ])

        os.makedirs(f'{os.getcwd()}/logs/Attention_Transfer', exist_ok=True)
        os.makedirs(self.save, exist_ok=True)
        self.logger_file = f'{os.getcwd()}/logs/Attention_Transfer/{self.name}.log'

        logging.basicConfig(
            filename=self.logger_file,
            format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.INFO,
        )

        logger.info(f'Experiment Arguments: {self.kwargs}')

        if self.wandb_monitor:
            wandb.init(project='Trailmet Attention_Transfer', name=self.name)
            wandb.config.update(self.kwargs)

    def compress_model(self):
        """Function to transfer knowledge from teacher to student."""
        # include teacher training options
        self.distill(
            self.teacher_model,
            self.student_model,
            self.dataloaders,
            **self.kwargs['DISTILL_ARGS'],
        )

    def distill(self, teacher_model, student_model, dataloaders, **kwargs):
        print('=====> TRAINING STUDENT NETWORK <=====')
        logger.info('=====> TRAINING STUDENT NETWORK <=====')

        self.register_hooks()

        test_only = kwargs.get('TEST_ONLY', False)
        weight_decay = kwargs.get('WEIGHT_DECAY', 0.0005)
        milestones = kwargs.get('MILESTONES', [82, 123])
        gamma = kwargs.get('GAMMA', 0.1)

        optimizer = torch.optim.SGD(
            student_model.parameters(),
            lr=self.lr,
            weight_decay=weight_decay,
            momentum=0.9,
        )
        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer,
                                                         milestones=milestones,
                                                         gamma=gamma,
                                                         verbose=False)

        criterion = self.criterion

        best_top1_acc = 0

        if test_only == False:
            epochs_list = []
            val_top1_acc_list = []
            val_top5_acc_list = []
            for epoch in range(self.epochs):
                t_loss = self.train_one_epoch(
                    teacher_model,
                    student_model,
                    dataloaders['train'],
                    criterion,
                    optimizer,
                    epoch,
                )

                valid_loss, valid_top1_acc, valid_top5_acc = self.test(
                    teacher_model, student_model, dataloaders['val'],
                    criterion, epoch)

                # use conditions for different schedulers e.g. ReduceLROnPlateau needs scheduler.step(v_loss)
                scheduler.step()

                is_best = False
                if valid_top1_acc > best_top1_acc:
                    best_top1_acc = valid_top1_acc
                    is_best = True

                save_checkpoint(
                    {
                        'epoch': epoch,
                        'state_dict': student_model.state_dict(),
                        'best_top1_acc': best_top1_acc,
                        'optimizer': optimizer.state_dict(),
                        'scheduler': scheduler.state_dict(),
                    },
                    is_best,
                    self.save,
                )

                if self.wandb_monitor:
                    wandb.log({'best_top1_acc': best_top1_acc})

                epochs_list.append(epoch)
                val_top1_acc_list.append(valid_top1_acc.cpu().numpy())
                val_top5_acc_list.append(valid_top5_acc.cpu().numpy())

                df_data = np.array([
                    epochs_list,
                    val_top1_acc_list,
                    val_top5_acc_list,
                ]).T
                df = pd.DataFrame(
                    df_data,
                    columns=[
                        'Epochs',
                        'Validation Top1',
                        'Validation Top5',
                    ],
                )
                df.to_csv(
                    f'{os.getcwd()}/logs/Attention_Transfer/{self.name}.csv',
                    index=False,
                )

    def train_one_epoch(self, teacher_model, student_model, dataloader,
                        loss_fn, optimizer, epoch):
        teacher_model.eval()
        student_model.train()

        batch_time = AverageMeter('Time', ':6.3f')
        data_time = AverageMeter('Data', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')

        end = time.time()

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Training student network Epoch [X] (X / X Steps) (batch time=X.Xs) (data time=X.Xs) (loss=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        for i, (images, labels) in enumerate(epoch_iterator):
            data_time.update(time.time() - end)
            images = images.to(self.device, dtype=torch.float)
            labels = labels.to(self.device)

            teacher_preds = teacher_model(images)
            student_preds = student_model(images)

            teacher_io_dict = self.forward_hook_manager_teacher.pop_io_dict()
            student_io_dict = self.forward_hook_manager_student.pop_io_dict()
            feature_map_pairs = []
            for teacher_layer, student_layer in zip(self.teacher_layer_names,
                                                    self.student_layer_names):
                feature_map_pairs.append((
                    teacher_io_dict[teacher_layer]['output'],
                    student_io_dict[student_layer]['output'],
                ))

            loss = loss_fn(teacher_preds, student_preds, feature_map_pairs,
                           labels)
            n = images.size(0)
            losses.update(loss.item(), n)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            batch_time.update(time.time() - end)
            end = time.time()

            epoch_iterator.set_description(
                'Training student network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(dataloader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                ))

            logger.info(
                'Training student network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (data time=%2.5fs) (loss=%2.5f)'
                % (
                    epoch,
                    (i + 1),
                    len(dataloader),
                    batch_time.val,
                    data_time.val,
                    losses.val,
                ))

            if self.wandb_monitor:
                wandb.log({
                    'train_loss': losses.val,
                })

        return losses.avg

    def test(self, teacher_model, student_model, dataloader, loss_fn, epoch):
        batch_time = AverageMeter('Time', ':6.3f')
        losses = AverageMeter('Loss', ':.4e')
        top1 = AverageMeter('Acc@1', ':6.2f')
        top5 = AverageMeter('Acc@5', ':6.2f')

        epoch_iterator = tqdm(
            dataloader,
            desc=
            'Validating student network Epoch [X] (X / X Steps) (batch time=X.Xs) (loss=X.X) (top1=X.X) (top5=X.X)',
            bar_format='{l_bar}{r_bar}',
            dynamic_ncols=True,
            disable=False,
        )

        teacher_model.eval()
        student_model.eval()

        with torch.no_grad():
            end = time.time()

            for i, (images, labels) in enumerate(epoch_iterator):
                images = images.to(self.device, dtype=torch.float)
                labels = labels.to(self.device)

                teacher_preds = teacher_model(images)
                student_preds = student_model(images)

                teacher_io_dict = self.forward_hook_manager_teacher.pop_io_dict(
                )
                student_io_dict = self.forward_hook_manager_student.pop_io_dict(
                )
                feature_map_pairs = []
                for teacher_layer, student_layer in zip(
                        self.teacher_layer_names, self.student_layer_names):
                    feature_map_pairs.append((
                        teacher_io_dict[teacher_layer]['output'],
                        student_io_dict[student_layer]['output'],
                    ))

                loss = loss_fn(teacher_preds, student_preds, feature_map_pairs,
                               labels)

                pred1, pred5 = accuracy(student_preds, labels, topk=(1, 5))

                n = images.size(0)
                losses.update(loss.item(), n)
                top1.update(pred1[0], n)
                top5.update(pred5[0], n)

                # measure elapsed time
                batch_time.update(time.time() - end)
                end = time.time()

                epoch_iterator.set_description(
                    'Validating student network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                logger.info(
                    'Validating student network Epoch [%d] (%d / %d Steps) (batch time=%2.5fs) (loss=%2.5f) (top1=%2.5f) (top5=%2.5f)'
                    % (
                        epoch,
                        (i + 1),
                        len(dataloader),
                        batch_time.val,
                        losses.val,
                        top1.val,
                        top5.val,
                    ))

                if self.wandb_monitor:
                    wandb.log({
                        'val_loss': losses.val,
                        'val_top1_acc': top1.val,
                        'val_top5_acc': top5.val,
                    })

            print(' * acc@1 {top1.avg:.3f} acc@5 {top5.avg:.3f}'.format(
                top1=top1, top5=top5))
        return losses.avg, top1.avg, top5.avg

    def criterion(self, teacher_preds, student_preds, feature_map_pairs,
                  labels):
        ce_loss = self.ce_loss(student_preds, labels)
        at_loss = self.at_loss(feature_map_pairs)
        return ce_loss + self.beta * at_loss

    def register_hooks(self):
        for layer in self.teacher_layer_names:
            self.forward_hook_manager_teacher.add_hook(self.teacher_model,
                                                       layer,
                                                       requires_input=False,
                                                       requires_output=True)

        for layer in self.student_layer_names:
            self.forward_hook_manager_student.add_hook(self.student_model,
                                                       layer,
                                                       requires_input=False,
                                                       requires_output=True)



# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import torch
import os
import yaml
import torch_pruning as tp
import numpy as np
from .benchmark import ModelBenchmark
from ..algorithms.prune.network_slimming import Network_Slimming
from ..algorithms.prune.pns import SlimPruner, ChannelRounding
from ..algorithms.prune.utils import cal_threshold_by_bn2d_weights


class TP_Prune:

    def __init__(
        self,
        method=None,
        prune_model=None,
        org_model=None,
        batch_size=64,
        input_size=32,
        device_name='cpu',
        root=None,
        yaml_name='resnet50_cifar100.yaml',
    ):
        self.root = root
        self.method = method
        self.prune_model = prune_model
        self.org_model = org_model
        self.batch_size = batch_size
        self.input_size = input_size
        self.device = torch.device(device_name)
        self.prune_model.to(self.device)
        self.org_model.to(self.device)

        # for network slimming
        if self.method == 'network_slimming':
            with open(os.path.join(self.root, yaml_name), 'r') as stream:
                data_loaded = yaml.safe_load(stream)
            data_loaded['schema_root'] = self.root
            slim = Network_Slimming(**data_loaded)
            self.pruner = SlimPruner(self.prune_model, slim.prune_schema)
            self.threshold = cal_threshold_by_bn2d_weights(
                [it.module for it in self.pruner.bn2d_modules.values()],
                slim.prune_ratio,
            )

    def prune(self):
        DG = tp.DependencyGraph().build_dependency(
            self.org_model,
            example_inputs=torch.randn(1, 3, self.input_size,
                                       self.input_size).to(self.device),
        )

        if self.method == 'chipnet':
            print('==> Chipnet method is selected for pruning')
            print('==> Pruning model started')

            for prune_model_modules, org_model_modules in zip(
                    self.prune_model.named_modules(),
                    self.org_model.named_modules()):
                if hasattr(prune_model_modules[1], 'pruned_zeta') and (
                        0 in prune_model_modules[1].pruned_zeta):
                    indices = (torch.where(prune_model_modules[1].pruned_zeta
                                           == 0)[0].cpu().tolist())
                    group = DG.get_pruning_group(
                        org_model_modules[1],
                        tp.prune_batchnorm_out_channels,
                        idxs=indices,
                    )

                    # # 3. prune all grouped layers that are coupled with model.conv1 (included).
                    if DG.check_pruning_group(
                            group):  # avoid full pruning, i.e., channels=0.
                        group.prune()

            print('==> Pruning model completed.')

        elif self.method == 'network_slimming':
            print('==> Network Slimming method is selected for pruning')
            print('==> Pruning model started')

            for name, module in self.org_model.named_modules():
                if 'model.' + name in list(self.pruner.bn2d_modules.keys()):
                    self.pruner.bn2d_modules['model.' + name].cal_keep_idxes(
                        self.threshold,
                        min_keep_ratio=0.02,
                        channel_rounding=ChannelRounding(
                            self.pruner.channel_rounding),
                    )

                    total_idxs = np.arange(
                        self.pruner.bn2d_modules['model.' +
                                                 name].module.num_features)
                    remove_idxs = np.setdiff1d(
                        total_idxs,
                        self.pruner.bn2d_modules['model.' + name].keep_idxes)

                    group = DG.get_pruning_group(
                        module,
                        tp.prune_batchnorm_out_channels,
                        idxs=remove_idxs.tolist(),
                    )

                    # # 3. prune all grouped layers that are coupled with model.conv1 (included).
                    if DG.check_pruning_group(
                            group):  # avoid full pruning, i.e., channels=0.
                        group.prune()

        return self.prune_model, self.org_model

    def benchmark_model(self, model):
        print('==> Model for benchmarking')
        print(model)
        print('==> Benchmarking model started')
        model_benchmark = ModelBenchmark(model, self.batch_size,
                                         self.input_size, self.device)
        model_benchmark.benchmark()



#!/usr/bin/env python

# MIT License
#
# Copyright (c) 2023 Transmute AI Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# importing the required packages
import numpy as np
from torch.utils.data.sampler import SubsetRandomSampler


class BaseDataset:
    """BaseDataset class is to be inherited by all dataset classes. All the
    generic attributes and methods will be a part of this class.

    Parameters
    ----------
        name (string): dataset name 'CIFAR10', 'CIFAR100', default=None.
        root (string): Root directory where ``cifar-10-batches-py`` exists or will be saved if download flag is set to
        True (default is None).
        train (bool, optional): If True, creates dataset from training set, otherwise
            creates from test set, default=None.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``. Default=None.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it, default=None.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again, default=True.
        split_types (list): the possible values of this parameter are "train", "test" and "val".
            If the split_type contains "val", then shuffle has to be True, default value is None.
        val_fraction (float): If float, should be between 0.0 and 1.0 and represent
        the proportion of the dataset to include in the val split.
        shuffle (bool): Whether or not to shuffle the data before splitting into val from train,
            default is True. If shuffle is true, there should be 'val' in split_types.
        random_seed (int): RandomState instance, default=None.
    """

    def __init__(
        self,
        name=None,
        root=None,
        transform=None,
        target_transform=None,
        download=True,
        split_types=None,
        val_fraction=0.2,
        shuffle=True,
        random_seed=42,
    ):
        self.name = name
        self.root = root
        self.transform = transform
        self.target_transform = target_transform
        self.download = download
        self.split_types = split_types
        self.val_fraction = val_fraction
        self.shuffle = shuffle
        self.random_seed = random_seed
        self.val_exists = False
        self.dataset_dict = {}

        # performing QC of the input parameters
        try:
            self.val_fraction = float(
                self.val_fraction)  # check for float type
        except ValueError:
            raise ValueError('val_fraction needs to be of float type')

        assert (self.val_fraction >= 0.0) and (
            self.val_fraction
            < 1.0), 'val_fraction should be in range [0.0, 1.0]'

        # check if split_types contain nothing beyond train, val and test
        assert (
            len(set(self.split_types) - set(['train', 'val', 'test'])) == 0
        ), 'Only train, val and test permitted as inputs in split_types'

        if self.shuffle:
            # ensure that val exists in split_types
            assert (
                'val' in self.split_types
            ), 'Required param val missing in the defined split types (split_type)'

    def stack_dataset(self):
        """
        Behavior:
            This function stacks the three dataset objects (train, val and test) in a single dictionary together with
            their samplers. For cases where the no validation set is explicitly available, the split is performed here.
        Returns:
            dataset_dict (dict): The keys of the dictionary are "train_datset", "val_dataset"
            and "test_dataset" and the values are object of pytorch CIFER10 dataset containing train,
            val and test respectively.
        """

        # defining the samplers
        self.dataset_dict['train_sampler'] = None
        self.dataset_dict['val_sampler'] = None
        self.dataset_dict['test_sampler'] = None

        if self.name in ['CIFAR10', 'CIFAR100', 'CHEST']:
            num_train = len(self.dataset_dict['train'])
            indices = list(range(num_train))
            split = int(np.floor(self.val_fraction * num_train))

            if self.shuffle:
                np.random.seed(self.random_seed)
                np.random.shuffle(indices)

            self.train_idx, self.valid_idx = indices[split:], indices[:split]
            train_sampler = SubsetRandomSampler(self.train_idx)
            valid_sampler = SubsetRandomSampler(self.valid_idx)
            self.dataset_dict['train_sampler'] = train_sampler
            self.dataset_dict['val_sampler'] = valid_sampler
        elif self.name == 'TCGA':
            self.train_idx, self.valid_idx = range(
                len(self.dataset_dict['train'])), range(
                    len(self.dataset_dict['val']))
            train_sampler = SubsetRandomSampler(self.train_idx)
            valid_sampler = SubsetRandomSampler(self.valid_idx)
            self.dataset_dict['train_sampler'] = train_sampler
            self.dataset_dict['val_sampler'] = valid_sampler

        return self.dataset_dict
